package org.netkernel.layer0.logging;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.*;
import org.netkernel.container.ILogger;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.util.Utils;

public class LogManager
{
	private static LogManager sInstance;
	private Map<ClassLoader,Object> mLoggers;
	private KernelLog mKernelLogger;
	private String mBasepath;
	private IHDSNode mPersistedLevels;
	
	public static final String DEFAULT_LOG="etc/system/LogConfig.xml";
	public static final String KERNEL_LOG_CONFIG_URL="etc/KernelLogConfig.xml";
	public static final String SAFE_KERNEL_LOG_CONFIG_URL="org/netkernel/layer0/logging/DefaultKernelLogConfig.xml";
	
	public static LogManager getInstance()
	{	return sInstance;
	}
	
	public LogManager(String aBasepath)
	{	mLoggers=new WeakHashMap<ClassLoader, Object>();
		mBasepath=aBasepath;
		sInstance=this;
		ResourceBundle kernelBundle=ResourceBundle.getBundle("org.netkernel.layer0.logging.messages");
		if (mBasepath!=null)
		{	try
			{	URL config=getClass().getClassLoader().getResource(KERNEL_LOG_CONFIG_URL);
				mKernelLogger=new KernelLog(config,kernelBundle,this);
			}
			catch (Exception e)
			{	System.out.println(Utils.throwableToString(e));
			}
		}
		
		if (mKernelLogger==null)
		{	try
			{	URL config=getClass().getClassLoader().getResource(SAFE_KERNEL_LOG_CONFIG_URL);
				mKernelLogger=new KernelLog(config,kernelBundle,this);
				
			}
			catch (Exception e2)
			{	//should not fail
			}
		}
	}
	
	public ILogger getKernelLogger()
	{	return mKernelLogger;
	}
	
	public String getLogfileBase() throws IOException
	{	
		String installPath=mBasepath;
		if (installPath==null)
		{	installPath=System.getProperty("java.io.tmpdir");
			File logDir=new File(installPath,"log/");
			logDir.mkdir();
		}
		return installPath;
	}
	
	public IApplicationLog getLog(String aId)
	{	IApplicationLog result=null;
		if (mKernelLogger.getId().equals(aId))
		{	result=mKernelLogger;
		}
		else
		{	synchronized(mLoggers)
			{	for (Object o : mLoggers.values())
				{	if (o instanceof IApplicationLog)
					{	IApplicationLog log=(IApplicationLog)o;
						if (log.getId().equals(aId))
						{	result=log;
							break;
						}
					}
				}
			}
		}
		return result;
	}
	
	public IApplicationLog getLog(IRequestScopeLevel aScope, String aResource)
	{	IApplicationLog result=null;
		synchronized(mLoggers)
		{	IRequestScopeLevel scope=aScope;
			do
			{	ClassLoader cl=scope.getSpace().getClassLoader();
				if (cl!=null)
				{	Object o=mLoggers.get(cl);
					if (o instanceof ApplicationLogImpl)
					{	result=(ApplicationLogImpl)o;
						if (result.isExpired())
						{	result=null;
							o=null;
						}
						else
						{	break;
						}
					}
					if (o==null)
					{	URL resourceURL = cl.getResource(aResource);
						if (resourceURL!=null)
						{	try
							{	result = new ApplicationLogImpl(resourceURL,mKernelLogger, this);
								if (mPersistedLevels!=null)
								{	IHDSNode logNode=mPersistedLevels.getNodes("/logs/log").filter(HDSPredicateFactory.namedChildStringEquals("id", result.getId())).getFirstNode();
									if (logNode!=null)
									{	readLogLevels(logNode.getFirstNode("levels"), result);
									}
								}
								mLoggers.put(cl, result);
							}
							catch (Exception e)
							{	mKernelLogger.log(ILogger.WARNING, this, "MSG_LR0_BAD_LOG",resourceURL,Utils.throwableToString(e));
								mLoggers.put(cl, Boolean.FALSE);
							}
							break;
						}
						else
						{	mLoggers.put(cl, Boolean.FALSE);
						}
					}
				}
				scope=scope.getParent();
			}
			while (scope!=null);
		}
		if (result==null)
		{	result=mKernelLogger;
			//TODO is there any warning needed when log not found?
			//if (aResource!=DEFAULT_LOG && !aResource.equals(DEFAULT_LOG)) {}
		}
		return result;		
	}
	
	public void setLevels(IHDSNode aState) throws Exception
	{	mPersistedLevels=aState;
		//update any logs already constructed
		for (IHDSNode logNode : mPersistedLevels.getNodes("/logs/log"))
		{	String id=logNode.getFirstValue("id").toString();
			IApplicationLog log=getLog(id);
			if (log!=null)
			{	readLogLevels(logNode.getFirstNode("levels"),log);
				String msg=mKernelLogger.format("MSG_LOG_LEVELS",toLogLevelsString(log));
				log.logRaw(ILogger.INFO, this, msg);
			}
		}		
	}

	
	public static Map<String,Integer> sLevelNameLookup = new HashMap<String,Integer>();
	static
	{	sLevelNameLookup.put("severe", ILogger.SEVERE);
		sLevelNameLookup.put("warning", ILogger.WARNING);
		sLevelNameLookup.put("info", ILogger.INFO);
		sLevelNameLookup.put("fine", ILogger.FINE);
		sLevelNameLookup.put("finer", ILogger.FINER);
		sLevelNameLookup.put("debug", ILogger.DEBUG);
		sLevelNameLookup.put("cache", ILogger.CACHE);
	}
	
	private void readLogLevels(IHDSNode aNode, IApplicationLog aLog) throws Exception
	{	
		for (IHDSNode n: aNode.getChildren())
		{	String name=n.getName();
			Integer level=sLevelNameLookup.get(name);
			if (level!=null)
			{	boolean shouldLog=Boolean.parseBoolean(n.getValue().toString());
				aLog.setShouldLog(level,null,shouldLog);
			}
		}
	}
	
	public IHDSNode getState()
	{	HDSBuilder b = new HDSBuilder();
		Set<String> instantiatedLogs = new HashSet();
		b.pushNode("logs");
		writeLogLevels(b, mKernelLogger);
		instantiatedLogs.add(mKernelLogger.getId());
		for (Object o: mLoggers.values())
		{	if (o instanceof IApplicationLog)
			{	IApplicationLog log=(IApplicationLog)o;
				if (!instantiatedLogs.contains(log.getId()))
				{	writeLogLevels(b,log);
					instantiatedLogs.add(log.getId());
				}
			}
		}
		//find any other logs that haven't been instantiated yet
		//and save their state too
		if (mPersistedLevels!=null)
		{	for (IHDSNode logNode : mPersistedLevels.getNodes("/logs/log"))
			{	String id=logNode.getFirstValue("id").toString();
				if (!instantiatedLogs.contains(id))
				{	b.pushNode("log");
					b.importChildren(logNode);
					b.popNode();
				}
			}
		}
		mPersistedLevels=b.getRoot();
		return mPersistedLevels;
	}
	
	private void writeLogLevels(HDSBuilder b, IApplicationLog aLog)
	{	b.pushNode("log");
		b.addNode("id", aLog.getId());
		b.addNode("name", aLog.getName());
		b.addNode("file", aLog.getStandardLogFile());
		b.pushNode("levels");
		b.addNode("severe",aLog.shouldLog(ILogger.SEVERE, null));
		b.addNode("warning",aLog.shouldLog(ILogger.WARNING, null));
		b.addNode("info",aLog.shouldLog(ILogger.INFO, null));
		b.addNode("fine",aLog.shouldLog(ILogger.FINE, null));
		b.addNode("finer",aLog.shouldLog(ILogger.FINER, null));
		b.addNode("debug",aLog.shouldLog(ILogger.DEBUG, null));
		b.addNode("cache",aLog.shouldLog(ILogger.CACHE, null));
		b.popNode();
		b.popNode();
	}
	
	private String toLogLevelsString(IApplicationLog aLog)
	{	StringBuilder sb=new StringBuilder(128);
		writeLogLevel(aLog,sb,"severe");
		writeLogLevel(aLog,sb,"warning");
		writeLogLevel(aLog,sb,"info");
		writeLogLevel(aLog,sb,"fine");
		writeLogLevel(aLog,sb,"finer");
		writeLogLevel(aLog,sb,"debug");
		writeLogLevel(aLog,sb,"cache");
		return sb.toString();
	}
	
	private void writeLogLevel(IApplicationLog aLog, StringBuilder sb, String aLevel)
	{
		sb.append(aLevel);
		sb.append('=');
		int level=sLevelNameLookup.get(aLevel);
		sb.append(aLog.shouldLog(level, null));
		sb.append(' ');
	}

}
